import torch.nn as nn
from typing import List, Sequence


def construct_mlp(
    input_dim: int,
    out_dim: int,
    hidden_layer_dims: Sequence[int],
    activation_layer: nn.Module = nn.ReLU(),
) -> nn.Sequential:
    """
    Construct a multi-linear perceptron (MLP). No non-linearity is *not* applied at the final layer.

    :param input_dim: the input dimension of the MLP.
    :param out_dim: the input dimension of the MLP.
    :param hidden_layer_dims: a list of zero or more integers indicating the dimensions of the
        hidden layers.
    :param activation_layer: the activation layer used between the input and hidden layers.

    :returns: a nn.Sequential with the constructed MLP.
    """
    layers: List[nn.Module] = []
    cur_hidden_dim = input_dim
    for hidden_layer_dim in hidden_layer_dims:
        layers.append(nn.Linear(cur_hidden_dim, hidden_layer_dim))
        layers.append(activation_layer)
        cur_hidden_dim = hidden_layer_dim
    layers.append(nn.Linear(cur_hidden_dim, out_dim))
    return nn.Sequential(*layers)



import unittest

from mi_module_zoo.mlp import construct_mlp
from mi_module_zoo.utils.testing import TensorSpec, generate_tensors, is_finite_tensor


class TestMLP(unittest.TestCase):
    def test_mlp(self):
        input_spec = {
            "input": TensorSpec("Batch", "input_dim"),
        }
        shapes, tensors = generate_tensors(
            input_spec, other_dims={"hidden_dim1", "hidden_dim2", "out_dim"}
        )

        hidden_layer_sizes = (shapes["hidden_dim1"], shapes["hidden_dim2"])

        # Test for 0..2 hidden layers
        for num_layers in range(3):
            with self.subTest(f"MLP with {num_layers} hidden layers."):
                mlp = construct_mlp(
                    input_dim=shapes["input_dim"],
                    out_dim=shapes["out_dim"],
                    hidden_layer_dims=hidden_layer_sizes[:num_layers],
                )

                out = mlp(tensors["input"])
                self.assertEqual(out.shape, (shapes["Batch"], shapes["out_dim"]))
                self.assertTrue(is_finite_tensor(out))



import torch
import unittest
from itertools import product
from typing import Iterator, NamedTuple

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from mi_module_zoo.relationaltransformerlayers import (
    RelationalTransformerDecoderLayer,
    RelationalTransformerEncoderLayer,
)
from mi_module_zoo.utils.testing import TensorSpec, generate_tensors, is_finite_tensor


class RelationalTransformerEncoderSettings(NamedTuple):
    use_dense_relations: bool
    use_sparse_edges: bool
    use_reverse_edges: bool
    use_mask: bool
    rezero_mode: Literal["off", "scalar", "vector", "scalar-tied"]
    normalisation_mode: Literal["off", "prenorm", "postnorm"]


class RelationalTransformerDecoderSettings(NamedTuple):
    use_dense_relations_enc: bool
    use_dense_relations_self: bool
    use_sparse_edges_enc: bool
    use_sparse_edges_self: bool
    use_reverse_edges: bool
    use_mask: bool
    rezero_mode: Literal["off", "scalar", "vector", "scalar-tied"]
    normalisation_mode: Literal["off", "prenorm", "postnorm"]


class TestRelationalTransformers(unittest.TestCase):
    def __relational_transformer_encoder_settings(
        self,
    ) -> Iterator[RelationalTransformerEncoderSettings]:
        """Yield all valid configurations of relational transformers."""
        for use_mask, use_dense_relations, use_sparse_edges in product((False, True), repeat=3):
            for rezero_mode in ("off", "scalar", "vector", "scalar-tied"):
                for norm_mode in ("off", "prenorm", "postnorm"):
                    if use_sparse_edges:
                        for use_reverse_edges in (False, True):
                            yield RelationalTransformerEncoderSettings(
                                use_dense_relations=use_dense_relations,
                                use_sparse_edges=True,
                                use_reverse_edges=use_reverse_edges,
                                use_mask=use_mask,
                                rezero_mode=rezero_mode,
                                normalisation_mode=norm_mode,
                            )
                    else:
                        yield RelationalTransformerEncoderSettings(
                            use_dense_relations=use_dense_relations,
                            use_sparse_edges=False,
                            use_reverse_edges=False,
                            use_mask=use_mask,
                            rezero_mode=rezero_mode,
                            normalisation_mode=norm_mode,
                        )

    def __relational_transformer_decoder_settings(
        self,
    ) -> Iterator[RelationalTransformerDecoderSettings]:
        """Yield all valid configurations of relational transformers."""
        for (
            use_mask,
            use_dense_relations_enc,
            use_dense_relations_self,
            use_sparse_edges_enc,
            use_sparse_edges_self,
        ) in product((False, True), repeat=5):
            for rezero_mode in ("off", "scalar", "vector", "scalar-tied"):
                for norm_mode in ("off", "prenorm", "postnorm"):
                    if use_sparse_edges_self:
                        for use_reverse_edges in (False, True):
                            yield RelationalTransformerDecoderSettings(
                                use_dense_relations_enc=use_dense_relations_enc,
                                use_dense_relations_self=use_dense_relations_self,
                                use_sparse_edges_self=True,
                                use_sparse_edges_enc=use_sparse_edges_enc,
                                use_reverse_edges=use_reverse_edges,
                                use_mask=use_mask,
                                rezero_mode=rezero_mode,
                                normalisation_mode=norm_mode,
                            )
                    else:
                        yield RelationalTransformerDecoderSettings(
                            use_dense_relations_enc=use_dense_relations_enc,
                            use_dense_relations_self=use_dense_relations_self,
                            use_sparse_edges_self=False,
                            use_sparse_edges_enc=use_sparse_edges_enc,
                            use_reverse_edges=False,
                            use_mask=use_mask,
                            rezero_mode=rezero_mode,
                            normalisation_mode=norm_mode,
                        )

    def test_relational_encoder(self):
        input_spec = {
            "src": TensorSpec("batch_size", "len", "D"),
            "mask": TensorSpec("batch_size", "len"),
            # Sparse relational edges
            "edges_batch_idx": TensorSpec("num_edges", indexed_dim="batch_size"),
            "edges_from_idx": TensorSpec("num_edges", indexed_dim="len"),
            "edges_to_idx": TensorSpec("num_edges", indexed_dim="len"),
            "edge_types": TensorSpec("num_edges", indexed_dim="num_edge_types"),
            # Dense relation edges
            "dense_relations_kq": TensorSpec("batch_size", "len", "len", "num_heads"),
            "dense_relations_kv": TensorSpec(
                "batch_size", "len", "len", "num_heads", "value_dimension"
            ),
        }
        shapes, tensors = generate_tensors(
            input_spec,
            other_dims={
                "key_query_dimension",
                "value_dimension",
                "num_heads",
                "num_edge_types",
                "dim_feedforward",
            },
        )
        print(shapes)
        sparse_edges = torch.stack(
            [tensors["edges_batch_idx"], tensors["edges_from_idx"], tensors["edges_to_idx"]], dim=-1
        )
        self.assertEqual(sparse_edges.shape, (shapes["num_edges"], 3))

        for rel_transformer_setting in self.__relational_transformer_encoder_settings():
            with self.subTest(**rel_transformer_setting._asdict()):
                print(rel_transformer_setting)
                encoder = RelationalTransformerEncoderLayer(
                    d_model=shapes["D"],
                    key_query_dimension=shapes["key_query_dimension"],
                    value_dimension=shapes["value_dimension"],
                    num_heads=shapes["num_heads"],
                    num_edge_types=shapes["num_edge_types"],
                    add_reverse_edges=rel_transformer_setting.use_reverse_edges,
                    dim_feedforward=shapes["dim_feedforward"],
                    use_edge_value_biases=rel_transformer_setting.use_sparse_edges,
                    rezero_mode=rel_transformer_setting.rezero_mode,
                    normalisation_mode=rel_transformer_setting.normalisation_mode,
                )

                if rel_transformer_setting.use_sparse_edges:
                    edges = sparse_edges
                    edge_types = tensors["edge_types"]
                else:
                    edges = torch.zeros((0, 3), dtype=torch.int64)
                    edge_types = torch.zeros(0, dtype=torch.int64)

                dense_relations_kq = (
                    tensors["dense_relations_kq"]
                    if rel_transformer_setting.use_dense_relations
                    else None
                )
                dense_relations_kv = (
                    tensors["dense_relations_kv"]
                    if rel_transformer_setting.use_dense_relations
                    else None
                )
                mask = tensors["mask"] > 0.5 if rel_transformer_setting.use_mask else None
                if mask is not None:
                    mask[:, 0] = False  # Enforce at least one element per set.

                out = encoder(
                    src=tensors["src"],
                    src_mask=mask,
                    edges=edges,
                    edge_types=edge_types,
                    dense_relations_kq=dense_relations_kq,
                    dense_relations_kv=dense_relations_kv,
                )

                self.assertEqual(out.shape, (shapes["batch_size"], shapes["len"], shapes["D"]))
                self.assertTrue(is_finite_tensor(out))

    def test_relational_decoder(self):
        """
        :param tgt: A [batch_size, seq_len, D] tensor.
        :param memory: A [batch_size, mem_len, D] tensor.
        :param tgt_mask: A [batch_size, seq_len] bool tensor
        :param self_edges: [num_self_edges, 3] each row has the form (batch_idx, source_idx, target_idx)
        :param self_edge_types: [num_self_edges] of integers from 0..num_self_edges
        :param encoder_edges: [num_enc_edges, 3] each row has the form (batch_idx, source_idx, target_idx)
        :param encoder_edge_types: [num_enc_edges] of integers from 0..num_enc_edges
        :param dense_self_relations_kq: [batch_size, seq_len, seq_len, num_heads]
        :param dense_self_relations_kv: [batch_size, seq_len, seq_len, num_heads, value_dimension]
        :param dense_encoder_relations_kq: [batch_size, seq_len, mem_len, num_heads]
        :param dense_encoder_relations_kv: [batch_size, seq_len, mem_len, num_heads, value_dimension]

        :return:  [batch_size, seq_len, H]
        """
        input_spec = {
            "tgt": TensorSpec("batch_size", "seq_len", "D"),
            "memory": TensorSpec("batch_size", "mem_len", "D"),
            "tgt_mask": TensorSpec("batch_size", "seq_len"),
            "memory_mask": TensorSpec("batch_size", "mem_len"),
            # Sparse self-relational edges
            "self_edges_batch_idx": TensorSpec("num_edges", indexed_dim="batch_size"),
            "self_edges_from_idx": TensorSpec("num_edges", indexed_dim="seq_len"),
            "self_edges_to_idx": TensorSpec("num_edges", indexed_dim="seq_len"),
            "self_edge_types": TensorSpec("num_edges", indexed_dim="num_self_edge_types"),
            # Sparse relational edges to memories
            "encoder_edges_batch_idx": TensorSpec("num_edges", indexed_dim="batch_size"),
            "encoder_edges_from_idx": TensorSpec("num_edges", indexed_dim="seq_len"),
            "encoder_edges_to_idx": TensorSpec("num_edges", indexed_dim="mem_len"),
            "encoder_edge_types": TensorSpec("num_edges", indexed_dim="num_edge_types_to_encoder"),
            # Dense relation edges
            "dense_self_relations_kq": TensorSpec("batch_size", "seq_len", "seq_len", "num_heads"),
            "dense_self_relations_kv": TensorSpec(
                "batch_size", "seq_len", "seq_len", "num_heads", "value_dimension"
            ),
            "dense_encoder_relations_kq": TensorSpec(
                "batch_size", "seq_len", "mem_len", "num_heads"
            ),
            "dense_encoder_relations_kv": TensorSpec(
                "batch_size", "seq_len", "mem_len", "num_heads", "value_dimension"
            ),
        }

        shapes, tensors = generate_tensors(
            input_spec,
            other_dims={
                "key_query_dimension",
                "value_dimension",
                "num_heads",
                "num_self_edge_types",
                "num_edge_types_to_encoder",
                "dim_feedforward",
            },
        )
        print(shapes)
        sparse_self_edges = torch.stack(
            [
                tensors["self_edges_batch_idx"],
                tensors["self_edges_from_idx"],
                tensors["self_edges_to_idx"],
            ],
            dim=-1,
        )
        self.assertEqual(sparse_self_edges.shape, (shapes["num_edges"], 3))

        sparse_encoder_edges = torch.stack(
            [
                tensors["encoder_edges_batch_idx"],
                tensors["encoder_edges_from_idx"],
                tensors["encoder_edges_to_idx"],
            ],
            dim=-1,
        )
        self.assertEqual(sparse_encoder_edges.shape, (shapes["num_edges"], 3))

        for rel_transformer_setting in self.__relational_transformer_decoder_settings():
            with self.subTest(**rel_transformer_setting._asdict()):
                print(rel_transformer_setting)
                decoder = RelationalTransformerDecoderLayer(
                    d_model=shapes["D"],
                    key_query_dimension=shapes["key_query_dimension"],
                    value_dimension=shapes["value_dimension"],
                    num_heads=shapes["num_heads"],
                    num_self_edge_types=shapes["num_self_edge_types"],
                    num_edge_types_to_encoder=shapes["num_edge_types_to_encoder"],
                    add_reverse_edges=rel_transformer_setting.use_reverse_edges,
                    dim_feedforward=shapes["dim_feedforward"],
                    use_edge_value_biases=rel_transformer_setting.use_sparse_edges_enc,
                    rezero_mode=rel_transformer_setting.rezero_mode,
                    normalisation_mode=rel_transformer_setting.normalisation_mode,
                )

                if rel_transformer_setting.use_sparse_edges_self:
                    self_edges = sparse_self_edges
                    self_edge_types = tensors["self_edge_types"]
                else:
                    self_edges = torch.zeros((0, 3), dtype=torch.int64)
                    self_edge_types = torch.zeros(0, dtype=torch.int64)

                if rel_transformer_setting.use_sparse_edges_enc:
                    encoder_edges = sparse_encoder_edges
                    encoder_edge_types = tensors["encoder_edge_types"]
                else:
                    encoder_edges = torch.zeros((0, 3), dtype=torch.int64)
                    encoder_edge_types = torch.zeros(0, dtype=torch.int64)

                dense_self_relations_kq = (
                    tensors["dense_self_relations_kq"]
                    if rel_transformer_setting.use_dense_relations_self
                    else None
                )
                dense_self_relations_kv = (
                    tensors["dense_self_relations_kv"]
                    if rel_transformer_setting.use_dense_relations_self
                    else None
                )
                dense_encoder_relations_kq = (
                    tensors["dense_encoder_relations_kq"]
                    if rel_transformer_setting.use_dense_relations_enc
                    else None
                )
                dense_encoder_relations_kv = (
                    tensors["dense_encoder_relations_kv"]
                    if rel_transformer_setting.use_dense_relations_enc
                    else None
                )
                tgt_mask = tensors["tgt_mask"] > 0.5 if rel_transformer_setting.use_mask else None
                if tgt_mask is not None:
                    # Enforce at least one element per set.
                    tgt_mask[:, 0] = False

                memory_mask = (
                    tensors["memory_mask"] > 0.5 if rel_transformer_setting.use_mask else None
                )
                if memory_mask is not None:
                    # Enforce at least one element per set.
                    memory_mask[:, 0] = False

                out = decoder(
                    tgt=tensors["tgt"],
                    memory=tensors["memory"],
                    tgt_mask=tgt_mask,
                    memory_mask=memory_mask,
                    self_edges=self_edges,
                    self_edge_types=self_edge_types,
                    encoder_edges=encoder_edges,
                    encoder_edge_types=encoder_edge_types,
                    dense_self_relations_kq=dense_self_relations_kq,
                    dense_self_relations_kv=dense_self_relations_kv,
                    dense_encoder_relations_kq=dense_encoder_relations_kq,
                    dense_encoder_relations_kv=dense_encoder_relations_kv,
                )

                self.assertEqual(out.shape, (shapes["batch_size"], shapes["seq_len"], shapes["D"]))
                self.assertTrue(is_finite_tensor(out))



import torch
import unittest
from itertools import product
from typing import Iterator, NamedTuple

from mi_module_zoo.relationalmultiheadattention import RelationalMultiheadAttention
from mi_module_zoo.utils.testing import TensorSpec, generate_tensors, is_finite_tensor


class RelationalMhaSettings(NamedTuple):
    use_dense_relations_kq: bool
    use_dense_relations_kv: bool

    use_sparse_edges: bool
    use_edge_value_biases: bool
    edge_attention_bias_is_scalar: bool

    use_mask: bool


class TestMultiheadAttention(unittest.TestCase):
    def __relational_mha_settings(self) -> Iterator[RelationalMhaSettings]:
        """Yield all valid configurations of relational transformers."""
        for use_mask, use_dense_relations_kq, use_dense_relations_kv, use_sparse_edges in product(
            (False, True), repeat=4
        ):
            if use_sparse_edges:
                for use_edge_value_biases, edge_attention_bias_is_scalar in product(
                    (False, True), repeat=2
                ):
                    yield RelationalMhaSettings(
                        use_dense_relations_kq=use_dense_relations_kq,
                        use_dense_relations_kv=use_dense_relations_kv,
                        use_sparse_edges=use_sparse_edges,
                        use_edge_value_biases=use_edge_value_biases,
                        edge_attention_bias_is_scalar=edge_attention_bias_is_scalar,
                        use_mask=use_mask,
                    )
            else:
                yield RelationalMhaSettings(
                    use_dense_relations_kq=use_dense_relations_kq,
                    use_dense_relations_kv=use_dense_relations_kv,
                    use_sparse_edges=use_sparse_edges,
                    use_edge_value_biases=False,
                    edge_attention_bias_is_scalar=False,
                    use_mask=use_mask,
                )

    def test_relational_mfa(self):
        input_spec = {
            "queries": TensorSpec("batch_size", "query_len", "num_heads", "D"),
            "keys": TensorSpec("batch_size", "key_len", "num_heads", "D"),
            "values": TensorSpec("batch_size", "key_len", "num_heads", "H"),
            "mask": TensorSpec("batch_size", "key_len"),
            # Sparse relational edges
            "edges_batch_idx": TensorSpec("num_edges", indexed_dim="batch_size"),
            "edges_from_idx": TensorSpec("num_edges", indexed_dim="query_len"),
            "edges_to_idx": TensorSpec("num_edges", indexed_dim="key_len"),
            "edge_types": TensorSpec("num_edges", indexed_dim="num_edge_types"),
            # Dense relation edges
            "dense_relations_kq": TensorSpec("batch_size", "query_len", "key_len", "num_heads"),
            "dense_relations_kv": TensorSpec(
                "batch_size", "query_len", "key_len", "num_heads", "H"
            ),
        }
        shapes, tensors = generate_tensors(input_spec, other_dims={"num_edge_types", "out_dim"})
        print(shapes)
        sparse_edges = torch.stack(
            [tensors["edges_batch_idx"], tensors["edges_from_idx"], tensors["edges_to_idx"]], dim=-1
        )
        self.assertEqual(sparse_edges.shape, (shapes["num_edges"], 3))

        for rmha_setting in self.__relational_mha_settings():
            with self.subTest(**rmha_setting._asdict()):
                print(rmha_setting)
                mha = RelationalMultiheadAttention(
                    num_heads=shapes["num_heads"],
                    num_edge_types=shapes["num_edge_types"],
                    key_query_dimension=shapes["D"],
                    value_dimension=shapes["H"],
                    output_dimension=shapes["out_dim"],
                    dropout_rate=0,
                    use_edge_value_biases=rmha_setting.use_edge_value_biases,
                    edge_attention_bias_is_scalar=rmha_setting.edge_attention_bias_is_scalar,
                )

                if rmha_setting.use_sparse_edges:
                    edges = sparse_edges
                    edge_types = tensors["edge_types"]
                else:
                    edges = torch.zeros((0, 3), dtype=torch.int64)
                    edge_types = torch.zeros(0, dtype=torch.int64)

                dense_relations_kq = (
                    tensors["dense_relations_kq"] if rmha_setting.use_dense_relations_kq else None
                )
                dense_relations_kv = (
                    tensors["dense_relations_kv"] if rmha_setting.use_dense_relations_kv else None
                )
                mask = tensors["mask"] > 0.5 if rmha_setting.use_mask else None
                if mask is not None:
                    mask[:, 0] = False  # Enforce at least one element per input.

                output = mha(
                    queries=tensors["queries"],
                    keys=tensors["keys"],
                    values=tensors["values"],
                    masked_elements=mask,
                    edges=edges,
                    edge_types=edge_types,
                    dense_relations_kq=dense_relations_kq,
                    dense_relations_kv=dense_relations_kv,
                )
                self.assertEqual(
                    output.shape, (shapes["batch_size"], shapes["query_len"], shapes["out_dim"])
                )
                self.assertTrue(is_finite_tensor(output))



import unittest
from itertools import product
from typing import Iterator, NamedTuple

from mi_module_zoo.settransformer import (
    ISAB,
    PMA,
    ElementwiseTransformType,
    MultiheadInitType,
    SetTransformer,
)
from mi_module_zoo.utils.testing import TensorSpec, generate_tensors, is_finite_tensor


class TestSetTransformer(unittest.TestCase):
    def test_isab(self):
        input_spec = {
            "x": TensorSpec("batch_size", "set_size", "per_head_embedding_dim", "num_heads"),
            "mask": TensorSpec("batch_size", "set_size"),
        }
        shapes, tensors = generate_tensors(
            input_spec,
            other_dims={
                "num_inducing_points",
            },
        )
        # Constraint in implementation
        embedding_dim = shapes["per_head_embedding_dim"] * shapes["num_heads"]

        mask = tensors["mask"] < 0.5
        mask[:, 0] = False  # Enforce at least one element per set.

        for mha_init, elementwise_transform_type, use_layer_norm in product(
            MultiheadInitType, ElementwiseTransformType, (False, True)
        ):
            with self.subTest(
                f"mha_init={mha_init}, elementwise_transform_type={elementwise_transform_type}, use_layer_norm={use_layer_norm}"
            ):
                isab = ISAB(
                    embedding_dim=embedding_dim,
                    num_heads=shapes["num_heads"],
                    num_inducing_points=shapes["num_inducing_points"],
                    multihead_init_type=mha_init,
                    use_layer_norm=use_layer_norm,
                    elementwise_transform_type=elementwise_transform_type,
                    dropout_rate=0.1,
                )
                x = tensors["x"]
                out = isab(
                    x=x.reshape(x.shape[0], x.shape[1], x.shape[2] * x.shape[3]),
                    mask=mask,
                )
                self.assertEqual(
                    out.shape, (shapes["batch_size"], shapes["set_size"], embedding_dim)
                )
                self.assertTrue(is_finite_tensor(out))

    def test_pma(self):
        input_spec = {
            "x": TensorSpec("batch_size", "set_size", "per_head_embedding_dim", "num_heads"),
            "mask": TensorSpec("batch_size", "set_size"),
        }
        shapes, tensors = generate_tensors(
            input_spec,
            other_dims={
                "num_seed_vectors",
            },
        )
        # Constraint in implementation
        embedding_dim = shapes["per_head_embedding_dim"] * shapes["num_heads"]
        mask = tensors["mask"] < 0.5
        mask[:, 0] = False  # Enforce at least one element per set.

        for (
            mha_init,
            elementwise_transform_type,
            use_layer_norm,
            use_elementwise_transform_pma,
        ) in product(
            MultiheadInitType,
            ElementwiseTransformType,
            (False, True),
            (False, True),
        ):
            with self.subTest(
                f"mha_init={mha_init}, elementwise_transform_type={elementwise_transform_type}, use_layer_norm={use_layer_norm}, use_elementwise_transform_pma={use_elementwise_transform_pma}"
            ):
                pma = PMA(
                    embedding_dim=embedding_dim,
                    num_heads=shapes["num_heads"],
                    num_seed_vectors=shapes["num_seed_vectors"],
                    multihead_init_type=mha_init,
                    use_layer_norm=use_layer_norm,
                    elementwise_transform_type=elementwise_transform_type,
                    use_elementwise_transform_pma=use_elementwise_transform_pma,
                    dropout_rate=0.1,
                )
                x = tensors["x"]
                out = pma(
                    x=x.reshape(x.shape[0], x.shape[1], x.shape[2] * x.shape[3]),
                    mask=mask,
                )

                self.assertEqual(
                    out.shape, (shapes["batch_size"], shapes["num_seed_vectors"], embedding_dim)
                )
                self.assertTrue(is_finite_tensor(out))

    class SetTransformerSettings(NamedTuple):
        use_transformer_embedding_dim: bool
        use_isab: bool
        multihead_init_type: MultiheadInitType
        use_layer_norm: bool
        elementwise_transform_type: ElementwiseTransformType
        use_elementwise_transform_pma: bool
        unit_seed_embeddings: bool

    def __set_transformer_settings(self) -> Iterator[SetTransformerSettings]:
        for mha_init, elementwise_transform_type in product(
            MultiheadInitType, ElementwiseTransformType
        ):
            for (
                use_transformer_embedding_dim,
                use_isab,
                use_layer_norm,
                use_elementwise_transform_pma,
                unit_seed_embeddings,
            ) in product((False, True), repeat=5):
                yield TestSetTransformer.SetTransformerSettings(
                    use_transformer_embedding_dim=use_transformer_embedding_dim,
                    use_isab=use_isab,
                    multihead_init_type=mha_init,
                    use_layer_norm=use_layer_norm,
                    elementwise_transform_type=elementwise_transform_type,
                    use_elementwise_transform_pma=use_elementwise_transform_pma,
                    unit_seed_embeddings=unit_seed_embeddings,
                )

    def test_set_transformer(self):
        for setting in self.__set_transformer_settings():
            with self.subTest(**setting._asdict()):
                if setting.use_transformer_embedding_dim:
                    input_spec = {
                        "x": TensorSpec("batch_size", "set_size", "input_embedding_dim"),
                        "mask": TensorSpec("batch_size", "set_size"),
                    }
                    other_dims = {"transformer_embedding_dim_per_head"}
                else:
                    input_spec = {
                        "x": TensorSpec(
                            "batch_size", "set_size", "input_embedding_dim_per_head", "num_heads"
                        ),
                        "mask": TensorSpec("batch_size", "set_size"),
                    }
                    other_dims = set()

                shapes, tensors = generate_tensors(
                    input_spec,
                    other_dims=other_dims
                    | {
                        "num_seed_vectors",
                        "set_embedding_dim",
                        "num_heads",
                        "num_blocks",
                        "num_inducing_points",
                        "set_embedding_dim",
                    },
                )
                print(shapes)
                mask = tensors["mask"] < 0.5
                mask[:, 0] = False  # Enforce at least one element per set.

                # Constraint in implementation
                if setting.use_transformer_embedding_dim:
                    transformer_embedding_dim = (
                        shapes["transformer_embedding_dim_per_head"] * shapes["num_heads"]
                    )
                    input_embedding_dim = shapes["input_embedding_dim"]
                    x = tensors["x"]
                else:
                    input_embedding_dim = (
                        shapes["input_embedding_dim_per_head"] * shapes["num_heads"]
                    )
                    x = tensors["x"]
                    x = x.reshape(x.shape[0], x.shape[1], x.shape[2] * x.shape[3])

                print(setting)
                num_seed_vectors = 1 if setting.unit_seed_embeddings else shapes["num_seed_vectors"]

                set_transformer = SetTransformer(
                    input_embedding_dim=input_embedding_dim,
                    set_embedding_dim=shapes["set_embedding_dim"],
                    transformer_embedding_dim=transformer_embedding_dim
                    if setting.use_transformer_embedding_dim
                    else None,
                    num_heads=shapes["num_heads"],
                    num_blocks=shapes["num_blocks"],
                    num_seed_vectors=num_seed_vectors,
                    use_isab=setting.use_isab,
                    num_inducing_points=shapes["num_inducing_points"] if setting.use_isab else None,
                    multihead_init_type=setting.multihead_init_type,
                    use_layer_norm=setting.use_layer_norm,
                    elementwise_transform_type=setting.elementwise_transform_type,
                    use_elementwise_transform_pma=setting.use_elementwise_transform_pma,
                )

                out = set_transformer(x, mask)
                self.assertEqual(
                    out.shape,
                    (
                        shapes["batch_size"],
                        shapes["set_embedding_dim"],
                    ),
                )
                self.assertTrue(is_finite_tensor(out))



import math
import torch
from enum import Enum
from torch import nn
from typing import Optional, cast

__all__ = ["MultiheadInitType", "ElementwiseTransformType", "SetTransformer", "ISAB", "PMA"]


class MultiheadInitType(Enum):
    XAVIER = 0
    KAIMING = 1


class ElementwiseTransformType(Enum):
    SINGLE = 0
    DOUBLE = 1


def _initialise_multihead(
    multihead: nn.MultiheadAttention, multihead_init_type: MultiheadInitType
) -> None:
    if multihead_init_type == MultiheadInitType.XAVIER:  # AIAYN and nn.MultiheadAttention default
        nn.init.xavier_uniform_(multihead.in_proj_weight)
        nn.init.constant_(multihead.in_proj_bias, 0.0)

        nn.init.kaiming_uniform_(multihead.out_proj.weight, a=math.sqrt(5))
        nn.init.constant_(multihead.out_proj.bias, 0.0)

    elif multihead_init_type == MultiheadInitType.KAIMING:
        # ST Implementation (nn.Linear) default
        nn.init.kaiming_uniform_(multihead.in_proj_weight, a=math.sqrt(5))
        in_proj_fan_in, _ = nn.init._calculate_fan_in_and_fan_out(multihead.in_proj_weight)
        in_proj_bound = 1 / math.sqrt(in_proj_fan_in)
        nn.init.uniform_(multihead.in_proj_bias, -in_proj_bound, in_proj_bound)

        nn.init.kaiming_uniform_(multihead.out_proj.weight, a=math.sqrt(5))
        out_proj_fan_in, _ = nn.init._calculate_fan_in_and_fan_out(multihead.out_proj.weight)
        out_proj_bound = 1 / math.sqrt(out_proj_fan_in)
        nn.init.uniform_(multihead.out_proj.bias, -out_proj_bound, out_proj_bound)
    else:
        raise ValueError(f"Unrecognized init type `{multihead_init_type}`.")


def _create_elementwise_transform(
    embedding_dim: int, elementwise_transform_type: ElementwiseTransformType
) -> nn.Sequential:
    if elementwise_transform_type == ElementwiseTransformType.SINGLE:  # ST Implementation default
        return nn.Sequential(
            nn.Linear(embedding_dim, embedding_dim),
            nn.ReLU(),
        )

    elif elementwise_transform_type == ElementwiseTransformType.DOUBLE:
        # AIAYN Implementation default
        return nn.Sequential(
            nn.Linear(embedding_dim, embedding_dim),
            nn.ReLU(),
            nn.Linear(embedding_dim, embedding_dim),
        )
    else:
        raise ValueError(f"Unrecognized elementwise transform type `{elementwise_transform_type}`.")


class SetTransformer(nn.Module):
    """
    The Set Transformer model https://arxiv.org/abs/1810.00825

    Generates an embedding from a set of features using several blocks of self attention
    and pooling by attention.
    """

    def __init__(
        self,
        input_embedding_dim: int,
        set_embedding_dim: int,
        transformer_embedding_dim: Optional[int] = None,
        num_heads: int = 1,
        num_blocks: int = 2,
        num_seed_vectors: int = 1,
        use_isab: bool = False,
        num_inducing_points: Optional[int] = None,
        multihead_init_type: MultiheadInitType = MultiheadInitType.XAVIER,
        use_layer_norm: bool = True,
        elementwise_transform_type: ElementwiseTransformType = ElementwiseTransformType.SINGLE,
        use_elementwise_transform_pma: bool = True,
        dropout_rate: float = 0.0,
    ):
        """
        Args:
            input_embedding_dim: Dimension of the input data, the embedded features.
            set_embedding_dim: Dimension of the output data, the set embedding.
            transformer_embedding_dim: Embedding dimension to be used in the set transformer blocks.
            num_heads: Number of heads in each multi-head attention block.
            num_blocks: Number of SABs in the model.
            num_seed_vectors: Number of seed vectors used in the pooling block (PMA).
            use_isab: Should ISAB blocks be used instead of SAB blocks.
            num_inducing_points: Number of inducing points.
            multihead_init_type: How linear layers in nn.MultiheadAttention are initialised. Valid options are "xavier" and "kaiming".
            use_layer_norm: Whether layer normalisation should be used in MAB blocks.
            elementwise_transform_type: What version of the elementwise transform (rFF) should be used. Valid options are "single" and "double".
            use_elementwise_transform_pma: Whether an elementwise transform (rFF) should be used in the PMA block.
        """
        super().__init__()
        self._transform_input_dimension = transformer_embedding_dim is not None
        if transformer_embedding_dim is not None:
            transformer_embedding_dim = cast(int, transformer_embedding_dim)
            assert (
                input_embedding_dim != transformer_embedding_dim
            ), "The linear transformation seems unnecessary since input_embedding_dim == transformer_embedding_dim"
            self._input_dimension_transform = nn.Linear(
                input_embedding_dim, transformer_embedding_dim
            )
        else:
            transformer_embedding_dim = input_embedding_dim

        if use_isab:
            if num_inducing_points is None:
                raise ValueError("Number of inducing points must be defined to use ISAB")
            self._sabs = nn.ModuleList(
                [
                    ISAB(
                        embedding_dim=transformer_embedding_dim,
                        num_heads=num_heads,
                        num_inducing_points=num_inducing_points,
                        multihead_init_type=multihead_init_type,
                        use_layer_norm=use_layer_norm,
                        elementwise_transform_type=elementwise_transform_type,
                        dropout_rate=dropout_rate,
                    )
                    for _ in range(num_blocks)
                ]
            )
        else:
            self._sabs = nn.ModuleList(
                [
                    SAB(
                        embedding_dim=transformer_embedding_dim,
                        num_heads=num_heads,
                        multihead_init_type=multihead_init_type,
                        use_layer_norm=use_layer_norm,
                        elementwise_transform_type=elementwise_transform_type,
                        dropout_rate=dropout_rate,
                    )
                    for _ in range(num_blocks)
                ]
            )

        self._pma = PMA(
            embedding_dim=transformer_embedding_dim,
            num_heads=num_heads,
            num_seed_vectors=num_seed_vectors,
            multihead_init_type=multihead_init_type,
            use_layer_norm=use_layer_norm,
            elementwise_transform_type=elementwise_transform_type,
            use_elementwise_transform_pma=use_elementwise_transform_pma,
            dropout_rate=dropout_rate,
        )
        self._output_dimension_transform = nn.Linear(
            transformer_embedding_dim * num_seed_vectors, set_embedding_dim
        )

    def forward(self, x: torch.Tensor, mask: Optional[torch.Tensor] = None) -> torch.Tensor:
        """
        Args:
            x: Embedded features tensor with shape [batch_size, set_size, input_embedding_dim].
            mask: Mask tensor with shape (batch_size, set_size), `True` values are masked
        Returns:
            set_embedding: Set embedding tensor with shape (batch_size, set_embedding_dim).
        """
        batch_size, _, _ = x.shape

        if self._transform_input_dimension:
            x = self._input_dimension_transform(
                x
            )  # Shape (batch_size, set_size, transformer_embedding_dim)
        for sab in self._sabs:
            x = sab(x, mask)  # Shape (batch_size, set_size, transformer_embedding_dim)
        x = self._pma(x, mask)  # Shape (batch_size, num_seed_vectors, transformer_embedding_dim)
        x = x.reshape(
            (batch_size, -1)
        )  # Shape (batch_size, num_seed_vectors * transformer_embedding_dim)
        set_embedding = self._output_dimension_transform(x)  # Shape (batch_size, set_embedding_dim)

        return set_embedding


class MAB(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        num_heads: int,
        multihead_init_type: MultiheadInitType,
        use_layer_norm: bool,
        elementwise_transform_type: ElementwiseTransformType,
        dropout_rate: float,
    ):
        """
        Multihead Attention Block of the Set Transformer model.

        :param embedding_dim: Dimension of the input data.
        :param num_heads: Number of heads.
        :param multihead_init_type: How linear layers in nn.MultiheadAttention are initialised.
        :param use_layer_norm: Whether layer normalisation should be used in MAB blocks.
        :param elementwise_transform_type: Elementwise transform (rFF) type used.
        :param dropout_rate: the percent of elements to dropout.
        """
        super().__init__()
        self._multihead = nn.MultiheadAttention(
            embedding_dim, num_heads, dropout=dropout_rate, batch_first=True
        )
        _initialise_multihead(self._multihead, multihead_init_type)

        self._use_layer_norm = use_layer_norm
        if self._use_layer_norm:
            self._layer_norm_1 = nn.LayerNorm(embedding_dim)
            self._layer_norm_2 = nn.LayerNorm(embedding_dim)

        self._elementwise_transform = _create_elementwise_transform(
            embedding_dim, elementwise_transform_type
        )

    def forward(
        self, query: torch.Tensor, key: torch.Tensor, key_mask: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """
        A multi-head attention block with keys==values. See Equation 6 of the Set Transformers paper.

        :param query: Query tensor with shape [batch_size, query_set_size, embedding_dim]
        :param key: Input tensor with shape [batch_size, key_set_size, embedding_dim] to be used as key and value.
        :param key_mask: Boolean mask tensor with shape [batch_size, key_set_size]. True values are masked.
                If None, nothing is masked.
        Returns:
            output: Tensor with shape [batch_size, query_set_size, embedding_dim].
        """
        x = (
            query
            + self._multihead(
                query=query, key=key, value=key, key_padding_mask=key_mask, need_weights=False
            )[0]
        )

        if self._use_layer_norm:
            x = self._layer_norm_1(x)

        x = x + self._elementwise_transform(x)

        if self._use_layer_norm:
            x = self._layer_norm_2(x)

        return x


class SAB(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        num_heads: int,
        multihead_init_type: MultiheadInitType,
        use_layer_norm: bool,
        elementwise_transform_type: ElementwiseTransformType,
        dropout_rate: float,
    ):
        """
        Self Attention Block of the Set Transformer model.

        Args:
        :param embedding_dim: Dimension of the input data.
        :param num_heads: Number of heads.
        :param multihead_init_type: How linear layers in nn.MultiheadAttention are initialised.
        :param use_layer_norm: Whether layer normalisation should be used in SAB blocks.
        :param elementwise_transform_type: Elementwise transform (rFF) type used.
        :param dropout_rate: the dropout rate
        """
        super().__init__()
        self._mab = MAB(
            embedding_dim=embedding_dim,
            num_heads=num_heads,
            multihead_init_type=multihead_init_type,
            use_layer_norm=use_layer_norm,
            elementwise_transform_type=elementwise_transform_type,
            dropout_rate=dropout_rate,
        )

    def forward(self, x: torch.Tensor, mask: Optional[torch.Tensor] = None) -> torch.Tensor:
        """
        :param x: Input tensor with shape [batch_size, set_size, embedding_dim] to be used as query, key, and value.
        :param mask: Boolean mask tensor with shape [batch_size, set_size], True values are masked.
                If None, everything is observed.
        Returns:
            output: Tensor with shape [batch_size, set_size, embedding_dim].
        """
        return self._mab(x, x, mask)


class PMA(nn.Module):
    """
    Pooling by Multihead Attention block of the Set Transformer model.
    """

    def __init__(
        self,
        embedding_dim: int,
        num_heads: int,
        num_seed_vectors: int,
        multihead_init_type: MultiheadInitType,
        use_layer_norm: bool,
        elementwise_transform_type: ElementwiseTransformType,
        use_elementwise_transform_pma: bool,
        dropout_rate: float,
    ):
        """
        :param embedding_dim: Dimension of the input data.
        :param num_heads: Number of heads.
        :param num_seed_vectors: Number of seed vectors.
        :param multihead_init_type: How linear layers in nn.MultiheadAttention are initialised.
        :param use_layer_norm: Whether layer normalisation should be used in MAB blocks.
        :param elementwise_transform_type: What version of the elementwise transform (rFF) should be used.
        :param use_elementwise_transform_pma: Whether an elementwise transform (rFF) should be used in the PMA block.
        """
        super().__init__()
        self._mab = MAB(
            embedding_dim=embedding_dim,
            num_heads=num_heads,
            multihead_init_type=multihead_init_type,
            use_layer_norm=use_layer_norm,
            elementwise_transform_type=elementwise_transform_type,
            dropout_rate=dropout_rate,
        )

        self._seed_vectors = nn.Parameter(
            torch.randn(1, num_seed_vectors, embedding_dim), requires_grad=True
        )
        nn.init.xavier_uniform_(self._seed_vectors)

        if use_elementwise_transform_pma:
            self._elementwise_transform = _create_elementwise_transform(
                embedding_dim, elementwise_transform_type
            )
        else:
            self._elementwise_transform = None

    def forward(self, x: torch.Tensor, mask: Optional[torch.Tensor] = None) -> torch.Tensor:
        """
        Seed vectors attend to the given values.
        The mask enforces that only the selected values are attended to in multihead attention.
        Args:
            x: Input tensor with shape [batch_size, set_size, embedding_dim] to be used as key and value.
            mask: Mask tensor with shape [batch_size, set_size], True for masked elements.
                If None, everything is observed.
        Returns:
            output: Attention output tensor with shape [batch_size, num_seed_vectors, embedding_dim].
        """
        if self._elementwise_transform:
            x = self._elementwise_transform(x)

        batch_size, _, _ = x.shape
        seed_vectors_repeated = self._seed_vectors.expand(batch_size, -1, -1)
        output = self._mab(seed_vectors_repeated, x, mask)

        return output


class ISAB(nn.Module):
    """
    Inducing-point self attention block. This reduces memory use and compute time from O(N^2) to O(NM)
    where N is the number of features and M is the number of inducing points.

    Reference: https://arxiv.org/pdf/1810.00825.pdf
    """

    def __init__(
        self,
        embedding_dim: int,
        num_heads: int,
        num_inducing_points: int,
        multihead_init_type: MultiheadInitType,
        use_layer_norm: bool,
        elementwise_transform_type: ElementwiseTransformType,
        dropout_rate: float,
    ):
        """
        Args:
            embedding_dim: Dimension of the input data.
            num_heads: Number of heads.
            num_inducing_points: Number of inducing points.
            multihead_init_type: How linear layers in nn.MultiheadAttention are initialised.
            use_layer_norm: Whether layer normalisation should be used in MAB blocks.
            elementwise_transform_type: What version of the elementwise transform (rFF) should be used.
        """
        super().__init__()
        self._mab1 = MAB(
            embedding_dim=embedding_dim,
            num_heads=num_heads,
            multihead_init_type=multihead_init_type,
            use_layer_norm=use_layer_norm,
            elementwise_transform_type=elementwise_transform_type,
            dropout_rate=dropout_rate,
        )
        self._mab2 = MAB(
            embedding_dim=embedding_dim,
            num_heads=num_heads,
            multihead_init_type=multihead_init_type,
            use_layer_norm=use_layer_norm,
            elementwise_transform_type=elementwise_transform_type,
            dropout_rate=dropout_rate,
        )
        self._inducing_points = nn.Parameter(
            torch.randn(1, num_inducing_points, embedding_dim), requires_grad=True
        )
        nn.init.xavier_uniform_(self._inducing_points)

    def forward(self, x: torch.Tensor, mask: Optional[torch.Tensor] = None) -> torch.Tensor:
        """
        The mask enforces that only the selected values are attended to in multihead attention,
        but the output is generated for all elements of x.
        Args:
            x: Input tensor with shape [batch_size, set_size, embedding_dim] to be used as query, key and value.
            mask: Mask tensor with shape [batch_size, set_size], True values are masked.
                If None, all elements are used is used.
        Returns:
            output: Attention output tensor with shape (batch_size, set_size, embedding_dim).
        """
        batch_size, _, _ = x.shape
        inducing_points = self._inducing_points.expand(
            batch_size, -1, -1
        )  # [batch_size, num_inducing_points, embedding_dim]
        y = self._mab1(
            query=inducing_points, key=x, key_mask=mask
        )  #  [batch_size, num_inducing_points, embedding_dim]
        return self._mab2(query=x, key=y)  # Shape [batch_size, set_size, embedding_dim]



import torch
from typing import Callable, Dict

try:
    from typing import Final
except ImportError:
    from typing_extensions import Final


def identity(x: torch.Tensor) -> torch.Tensor:
    return x


ACTIVATION_FNS: Final[Dict[str, Callable[[torch.Tensor], torch.Tensor]]] = {
    "celu": torch.nn.functional.celu,
    "elu": torch.nn.functional.elu,
    "gelu": torch.nn.functional.gelu,
    "hardswish": torch.nn.functional.hardswish,
    "hardtanh": torch.nn.functional.hardtanh,
    "identity": identity,
    "leaky_relu": torch.nn.functional.leaky_relu,
    "logsigmoid": torch.nn.functional.logsigmoid,
    "log_sigmoid": torch.nn.functional.logsigmoid,
    "none": identity,
    "relu": torch.relu,
    "relu6": torch.nn.functional.relu6,
    "rrelu": torch.nn.functional.rrelu,
    "sigmoid": torch.sigmoid,
    "selu": torch.nn.functional.selu,
    "silu": torch.nn.functional.silu,
    "softplus": torch.nn.functional.softplus,
    "swish": torch.nn.functional.silu,
    "tanh": torch.tanh,
}


def get_activation_fn(activation: str) -> Callable[[torch.Tensor], torch.Tensor]:
    """
    Get an activation function by name.
    :param activation: the name of the activation function.

    :returns: an activation function
    """
    activation = activation.lower()
    if activation not in ACTIVATION_FNS:
        raise RuntimeError(
            "Supported activations are `{}`, not {}".format(ACTIVATION_FNS.keys(), activation)
        )

    return ACTIVATION_FNS[activation]



import math
import torch
from torch import nn
from typing import Optional


class RelationalMultiheadAttention(nn.Module):
    """
    A relational multihead implementation supporting two variations of using additional
    relationship information between tokens:

    * If no edge information is passed in in .forward(), this behaves like a standard
      multi-head self-attention.
    * If edges are present and edge_attention_bias_is_scalar=False,
      and use_edge_value_biases=True is set, this implements
          Eqs. (3) and (4) of
            Shaw, Peter, Jakob Uszkoreit, and Ashish Vaswani. "Self-attention with relative position representations."
            In ACL 2018.  https://www.aclweb.org/anthology/N18-2074/
        and
          Eq. (2) of
            Wang, Bailin, et al. "RAT-SQL: Relation-aware schema encoding and linking for text-to-SQL parsers."
            In ICML 2020.  https://arxiv.org/pdf/1911.04942.pdf
    * If edges are present and edge_attention_bias_is_scalar=True,
      and use_edge_value_biases=False is set, this implements Sect. 3.1 of
        Hellendoorn, Vincent J., et al. "Global relational modules of source code."
        In ICLR 2020. https://openreview.net/pdf?id=B1lnbRNtwr
    """

    def __init__(
        self,
        *,
        num_heads: int,
        num_edge_types: int,
        key_query_dimension: int,
        value_dimension: int,
        output_dimension: int,
        dropout_rate: float,
        use_edge_value_biases: bool = False,
        edge_attention_bias_is_scalar: bool = False,
    ):
        super().__init__()
        assert num_heads > 0
        assert 0 <= dropout_rate < 1

        self._num_heads = num_heads
        self._value_dim = value_dimension
        self._key_query_dim = key_query_dimension

        self._dropout_layer = nn.Dropout(p=dropout_rate)
        self._scaling = key_query_dimension ** -0.5
        self._out_proj = nn.Linear(value_dimension * num_heads, output_dimension, bias=False)

        self._use_edge_value_biases = use_edge_value_biases
        self._edge_attention_bias_is_scalar = edge_attention_bias_is_scalar

        if self._edge_attention_bias_is_scalar:
            edge_attention_bias_dim = num_heads
        else:
            edge_attention_bias_dim = num_heads * key_query_dimension
        self._edge_attention_biases = nn.Embedding(
            num_embeddings=num_edge_types, embedding_dim=edge_attention_bias_dim
        )

        if self._use_edge_value_biases:
            self._edge_value_biases = nn.Embedding(
                num_embeddings=num_edge_types, embedding_dim=num_heads * value_dimension
            )

    @property
    def num_edge_types(self) -> int:
        return self._edge_attention_biases.num_embeddings

    @property
    def num_heads(self) -> int:
        return self._num_heads

    @property
    def value_dim(self) -> int:
        return self._value_dim

    @property
    def key_query_dim(self) -> int:
        return self._key_query_dim

    def forward(
        self,
        *,
        queries: torch.Tensor,
        keys: torch.Tensor,
        values: torch.Tensor,
        masked_elements: Optional[torch.Tensor],
        edges: torch.Tensor,
        edge_types: torch.Tensor,
        dense_relations_kq: Optional[torch.Tensor] = None,
        dense_relations_kv: Optional[torch.Tensor] = None,
    ):
        """
        :param queries: [batch_size, query_len, D]
        :param keys: [batch_size, key_len, D]
        :param values: [batch_size, key_len, H]
        :param masked_elements: [batch_size, key_len]  `True` values are those
            that should be masked (no attention paid).
        :param edges: [num_edges, 3] each row has the form (batch_idx, source_idx, target_idx)
        :param edge_types: [num_edges] of integers from 0..num_edge_types
        :param dense_relations_kq: [batch_size, query_len, key_len, num_heads]
        :param dense_relations_kv: [batch_size, query_len, key_len, num_heads, value_dimension]
        :return:  [batch_size, seq_size, H]
        """
        edge_sample_ids = edges[:, 0]
        edge_sources = edges[:, 1]
        edge_targets = edges[:, 2]

        # Standard dot-attention: Here, we compute
        #    e_bijk = (in_bi * W_Q^k) * (in_bj * W_K^k)^T
        # i.e., the inner product of the query-projection of token in_bi and key-projection of token in_bj,
        # where b is the ID of the sample in the batch, i, j are token IDs, and k is the ID of a head.
        raw_attention_scores = torch.einsum(
            "bkhd,bqhd->bqkh", keys, queries
        )  # [B, query_len, memory_len, num_heads]

        if edge_sample_ids.shape[0] > 0:
            attention_scores = self._add_edge_attention_scores(
                edge_sample_ids,
                edge_sources,
                edge_targets,
                edge_types,
                keys,
                queries,
                raw_attention_scores,
            )
        else:
            attention_scores = raw_attention_scores

        if dense_relations_kq is not None:
            attention_scores = attention_scores + dense_relations_kq

        attention_scores = attention_scores.transpose(2, 3)  # [B, query_len, num_heads, key_len]

        if masked_elements is not None:
            attention_scores.masked_fill_(masked_elements.unsqueeze(1).unsqueeze(1), -math.inf)

        attention_probs = torch.softmax(attention_scores, dim=-1)
        attention_probs = self._dropout_layer(attention_probs)

        multiheaded_weighted_value_sum = torch.einsum(
            "blhq,bqhd->blhd", attention_probs, values
        )  # [B, query_len, num_heads, value_dim]

        if self._use_edge_value_biases:
            multiheaded_weighted_value_sum = self._add_edge_value_biases(
                edge_sample_ids,
                edge_sources,
                edge_targets,
                edge_types,
                attention_probs,
                multiheaded_weighted_value_sum,
            )
        if dense_relations_kv is not None:
            value_biases = torch.einsum("bqhk,bqkhd->bqhd", attention_probs, dense_relations_kv)
            multiheaded_weighted_value_sum = multiheaded_weighted_value_sum + value_biases

        attention_output = multiheaded_weighted_value_sum.reshape(
            (
                multiheaded_weighted_value_sum.shape[0],
                multiheaded_weighted_value_sum.shape[1],
                -1,
            )
        )  # [B, query_len, num_heads * value_dim]
        return self._out_proj(attention_output)  # [B, query_len, out_dim]

    def _add_edge_attention_scores(
        self,
        edge_sample_ids,
        edge_sources,
        edge_targets,
        edge_types,
        keys,
        queries,
        raw_attention_scores,
    ):
        # We compute (sparse, per existing edge) additional bias scores e'_bijk:
        edge_bias_scores = self._compute_edge_bias_scores(
            edge_sample_ids, edge_sources, edge_targets, edge_types, keys, queries
        )

        # We add the e'_bijk (where present) to e_bijk. This should be a simple +=, but
        # that doesn't accumulate if we have several entries to add to e_bij. Hence we use
        # index_put_, which in turn requires a contiguous Tensor memory layout, and so we need
        # to establish that first:
        attention_scores = raw_attention_scores.contiguous()
        edge_sample_indices = edge_sample_ids
        edge_query_indices = edge_sources
        edge_key_indices = edge_targets
        attention_scores.index_put_(
            indices=(edge_sample_indices, edge_query_indices, edge_key_indices),
            values=edge_bias_scores,
            accumulate=True,
        )

        return attention_scores

    def _compute_edge_bias_scores(
        self, edge_sample_ids, edge_sources, edge_targets, edge_types, keys, queries
    ):
        # We will compute additional e'_bihj which will be added onto the standard attention scores:
        attention_biases = self._edge_attention_biases(edge_types)

        if self._edge_attention_bias_is_scalar:
            # Compute e'_bijk = \sum_d (bias_bijk * (in_bj * W_K^k))_d
            # This is the GREAT model. Note two things:
            #  (1) This is defined on the _key_ representation, not the _query_ repr.
            #  (2) Because bias_bijk is a scalar, this is essentially just scaling
            #      (in_bj * W_K^k) and then summing.
            edge_bias_scores = torch.einsum(
                "eh,ehd->eh",
                attention_biases,  # [num_edges, num_heads]
                keys[edge_sample_ids, edge_targets],  # [num_edges, num_heads, key_dim]
            )  # [num_edges, num_head]
        else:
            # Compute e'_bijk = (in_bj * W_Q^k) * bias_bijk^T
            # This is the Relative Position Representations / RAT-SQL variant. Note that this
            # is defined using the query representation, not the key repr.
            edge_bias_scores = torch.einsum(
                "ehd,ehd->eh",
                attention_biases.reshape((-1, self._num_heads, self._key_query_dim)),
                # [num_edges, num_heads, key_dim]
                keys[edge_sample_ids, edge_targets],  # [num_edges, num_heads, key_dim]
            )  # [num_edges, num_head]
        return edge_bias_scores

    def _add_edge_value_biases(
        self,
        edge_sample_ids,
        edge_sources,
        edge_targets,
        edge_types,
        attention_probs,
        multiheaded_weighted_value_sum,
    ):
        edge_sample_indices = edge_sample_ids
        edge_query_indices = edge_sources

        value_biases_shape = (
            edge_sample_ids.shape[0],
            self._num_heads,
            self._value_dim,
        )
        value_bias_per_edge = attention_probs[
            edge_sample_ids, edge_sources, :, edge_targets
        ].unsqueeze(-1) * self._edge_value_biases(edge_types).reshape(
            value_biases_shape
        )  # [num_edges, num_heads, value_dim]

        biased_weighted_value_sum = multiheaded_weighted_value_sum.contiguous()
        biased_weighted_value_sum.index_put_(
            indices=(edge_sample_indices, edge_query_indices),
            values=value_bias_per_edge,
            accumulate=True,
        )
        return biased_weighted_value_sum



import unittest

from mi_module_zoo.utils.activation import ACTIVATION_FNS
from mi_module_zoo.utils.testing import TensorSpec, generate_tensors, is_finite_tensor


class TestActivations(unittest.TestCase):
    def test_activation(self):
        input_spec = {
            "input1": TensorSpec("A", "B"),
            "input2": TensorSpec(
                "C",
            ),
        }
        shapes, tensors = generate_tensors(
            input_spec,
        )

        for activation_name, activation_fn in ACTIVATION_FNS.items():
            with self.subTest(f"Activation {activation_name}."):
                print(activation_name)
                for input in (tensors["input1"], tensors["input2"]):
                    result = activation_fn(input)
                    self.assertEqual(result.shape, input.shape)
                    self.assertTrue(is_finite_tensor(result))



import random
import torch
from typing import Collection, Dict, Optional, Set, Tuple, Union

try:
    from typing import Final
except ImportError:
    from typing_extensions import Final

__all__ = ["TensorSpec", "generate_tensors"]


class TensorSpec:
    def __init__(self, *dims: Union[str, int], indexed_dim: Optional[str] = None):
        """
        A symbolic specification of a tensor.

        :param dims: a variable-sized list of dimensions. Symbolic names (strings) or concrete
            sizes (integers) can be given.
        :param indexed_dim: Optional. If the tensor indexes another dimension, then the symbolic
             name of that dimension. This implies that the tensor is a torch.int64 tensor with
             values in [0, size(indexed_dim)).
        """
        self.dim: Final = dims
        self.indexed_dim: Final = indexed_dim


_PRIMES: Final = [
    3,
    5,
    7,
    11,
    13,
    17,
    19,
    23,
    29,
    31,
    37,
    41,
    43,
    47,
    53,
    59,
    61,
    67,
    71,
    73,
    79,
    83,
    89,
    97,
    101,
    103,
    107,
    109,
    113,
    127,
    131,
    137,
    139,
]


def generate_tensors(
    tensor_specs: Dict[str, TensorSpec],
    other_dims: Collection[str] = frozenset(),
    rng: random.Random = random,
) -> Tuple[Dict[str, int], Dict[str, torch.Tensor]]:
    """
    Generate random tensors and assign sizes to dimension based on the input specifications,
        to be used for testing.

    All dimensions will be assigned a different concrete shape which is a prime number. Prime
        shape sizes provide some reassurance that each dimension cannot be confused with another
        one, e.g. due to reshaping.

    :param tensor_specs: a dictionary of tensor names with their associated `TensorSpec`s.
    :param other_dims: a set of named dimensions that should be assigned a concrete size.
    :param rng: the random number generator to be used. Default the `random` package.
    :returns: a tuple with two elements
        concrete_dim_sizes: the concrete size of the dimensions in `tensor_specs` and `other_dims`
        generated_tensors: a dictionary with the concrete random tensors defined in `tensor_specs`.
    """

    # Collect all symbolic dims
    dims: Set[str] = set()
    for spec in tensor_specs.values():
        dims.update(s for s in spec.dim if isinstance(s, str))
    dims.update(other_dims)

    # Assign random (first len(dims) prime) sizes to dims
    shape_sizes = dict(zip(dims, rng.sample(_PRIMES[: len(dims)], k=len(dims))))

    # Generate the random test tensors
    out_tensors = {}
    for tensor_name, tensor_spec in tensor_specs.items():
        shape = tuple(shape_sizes[dim] if isinstance(dim, str) else dim for dim in tensor_spec.dim)
        if tensor_spec.indexed_dim is None:
            tensor = torch.randn(shape)
        else:
            assert (
                tensor_spec.indexed_dim in dims
            ), f"Indexed dimension {tensor_spec.indexed_dim} was not specified in input specs."
            tensor = torch.randint(shape_sizes[tensor_spec.indexed_dim], shape)

        out_tensors[tensor_name] = tensor

    return shape_sizes, out_tensors


def is_finite_tensor(tensor: torch.Tensor) -> bool:
    return not torch.any(torch.isnan(tensor) | torch.isinf(tensor))



import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), "README.md")) as f:
    long_description = f.read()

setup(
    name="mi-module-zoo",
    packages=find_packages(),
    license="MIT",
    package_dir={"mi-module-zoo": "mi_module_zoo"},
    test_suite="tests",
    python_requires=">=3.6.1",
    description="Reusable PyTorch Modules",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Machine Intelligence",
    package_data={"ptgnn": ["py.typed"]},
    install_requires=[
        "torch>=1.9.1",
        "typing-extensions",
    ],
    extras_require={"dev": ["black", "isort", "pre-commit", "numpy"], "aml": ["azureml"]},
    setup_requires=["setuptools_scm"],
    url="https://github.com/microsoft/mi-module-zoo/",
    project_urls={
        "Bug Tracker": "https://github.com/microsoft/mi-module-zoo/issues",
        # "Documentation": "https://github.com/microsoft/mi-module-zoo/tree/master/docs",
        "Source Code": "https://github.com/microsoft/mi-module-zoo",
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Libraries",
        "Typing :: Typed",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    zip_safe=False,
)



import numpy as np
import random
import torch


def set_seed(seed: int) -> None:
    """
    Set the seed for most common sources of randomness.
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.benchmark = True



import torch
from torch import nn
from typing import Callable, Optional, Tuple, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from mi_module_zoo.relationalmultiheadattention import RelationalMultiheadAttention
from mi_module_zoo.utils.activation import get_activation_fn


class RelationalTransformerEncoderLayer(nn.Module):
    def __init__(
        self,
        d_model: int,
        key_query_dimension: int,
        value_dimension: int,
        num_heads: int,
        num_edge_types: int,
        add_reverse_edges: bool = True,
        dim_feedforward: int = 2048,
        dropout_rate: float = 0.1,
        activation: str = "relu",
        use_edge_value_biases: bool = False,
        edge_attention_bias_is_scalar: bool = False,
        rezero_mode: Literal["off", "scalar", "vector", "scalar-tied"] = "off",
        normalisation_mode: Literal["off", "prenorm", "postnorm"] = "postnorm",
    ):
        """
        Args:
            - rezero_mode: Three different modes are supported:
                * "off": No ReZero use.
                * "scalar": Sublayers (attention / fully connected) are scaled by a single scalar, i.e.,
                  \alpha is a scalar in the following:
                    x' = x + \alpha * SelfAtt(x)
                    x'' = x' + \alpha * Boom(x')
                    return x''
                  See https://arxiv.org/pdf/2003.04887.pdf.
                * "vector": Sublayers (attention / fully connected) are scaled by one value per dim, i.e.,
                  \alpha is a vector in the following:
                    x' = x + \alpha * SelfAtt(x)
                    x'' = x' + \alpha * Boom(x')
                    return x''
                  See https://arxiv.org/pdf/2103.17239.pdf.
            - normalisation_mode: Three different modes are supported:
                * "off": use no layer norm at all. Likely to diverge without using rezero as well.
                * "prenorm": Normalise values before each sublayer (attention / fully connected):
                    x' = x + SelfAtt(LN(x))
                    x'' = x' + Boom(LN(x'))
                    return x''
                * "postnorm": Normalise values after each sublayer:
                    x' = LN(x + SelfAtt(x))
                    x'' = LN(x' + Boom(x))
                    return x''
        """
        super(RelationalTransformerEncoderLayer, self).__init__()
        assert 0 <= dropout_rate < 1

        self.self_attn = RelationalMultiheadAttention(
            num_heads=num_heads,
            output_dimension=d_model,
            dropout_rate=dropout_rate,
            num_edge_types=2 * num_edge_types if add_reverse_edges else num_edge_types,
            key_query_dimension=key_query_dimension,
            value_dimension=value_dimension,
            use_edge_value_biases=use_edge_value_biases,
            edge_attention_bias_is_scalar=edge_attention_bias_is_scalar,
        )
        # Implementation of Feedforward model
        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.dropout = nn.Dropout(dropout_rate)
        self.linear2 = nn.Linear(dim_feedforward, d_model)

        self._selfatt_head_transforms = nn.Linear(
            in_features=d_model,
            out_features=num_heads * (2 * key_query_dimension + value_dimension),
            bias=False,
        )

        self._normalisation_mode = normalisation_mode
        if normalisation_mode in ("prenorm", "postnorm"):
            self.norm1: Optional[nn.LayerNorm] = nn.LayerNorm(d_model)
            self.norm2: Optional[nn.LayerNorm] = nn.LayerNorm(d_model)
        elif normalisation_mode == "off":
            self.norm1 = None
            self.norm2 = None
        else:
            raise ValueError(f"Unrecognized normalization mode `{normalisation_mode}`.")

        self.dropout1 = nn.Dropout(dropout_rate)
        self.dropout2 = nn.Dropout(dropout_rate)

        self.activation = get_activation_fn(activation)

        self._rezero_mode = rezero_mode
        if rezero_mode == "off":
            self._alpha1: Union[float, torch.Tensor] = 1.0
            self._alpha2: Union[float, torch.Tensor] = 1.0
        elif rezero_mode == "scalar":
            self._alpha1 = nn.Parameter(torch.tensor(0.0))
            self._alpha2 = nn.Parameter(torch.tensor(0.0))
        elif rezero_mode == "scalar-tied":
            # The original ReZero setting: https://github.com/majumderb/rezero/blob/e2c94a825c5564217e8cf4d75a28d59cab1d7029/rezero/transformer/rztx.py#L47
            self._alpha1 = nn.Parameter(torch.tensor(0.0))
            self._alpha2 = self._alpha1
        elif rezero_mode == "vector":
            self._alpha1 = nn.Parameter(torch.zeros(size=(d_model,)))
            self._alpha2 = nn.Parameter(torch.zeros(size=(d_model,)))
        else:
            raise ValueError(f"Unrecognized rezero mode `{rezero_mode}`.")

        self._num_edge_types = num_edge_types
        self._add_reverse_edges = add_reverse_edges

    def _compute_qkv(
        self, input_seq_states: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        keys_queries_values = self._selfatt_head_transforms(input_seq_states).reshape(
            input_seq_states.shape[0],
            input_seq_states.shape[1],
            self.self_attn._num_heads,
            -1,
        )
        queries, keys, values = torch.split(
            keys_queries_values,
            split_size_or_sections=[
                self.self_attn._key_query_dim,
                self.self_attn._key_query_dim,
                self.self_attn._value_dim,
            ],
            dim=-1,
        )  # [B, query_len, num_heads, key_dim], [B, memory_len, num_heads, key_dim], [B, memory_len, num_heads, value_dim]

        return queries, keys, values

    def forward(
        self,
        src: torch.Tensor,
        src_mask: torch.Tensor,
        edges: torch.Tensor,
        edge_types: torch.Tensor,
        dense_relations_kq: Optional[torch.Tensor] = None,
        dense_relations_kv: Optional[torch.Tensor] = None,
        post_self_att_hook: Optional[
            Callable[[torch.Tensor, Union[float, torch.Tensor]], torch.Tensor]
        ] = None,
    ):
        """
        :param src: A [batch_size, seq_len, D] tensor.
        :param src_mask: A [batch_size, seq_len] bool tensor.  `True` values are those
            that should be masked (no attention paid).
        :param edges: [num_edges, 3] each row has the form (batch_idx, source_idx, target_idx)
        :param edge_types: [num_edges] of integers from 0..num_edge_types
        :param dense_relations_kq: [batch_size, seq_len, seq_len, num_heads]
        :param dense_relations_kv: [batch_size, seq_len, seq_len, num_heads, value_dimension]
        :return:  [batch_size, seq_len, D]
        """
        # --- Sublayer 1: Self-Attention:
        attn_input = src
        if self._normalisation_mode == "prenorm":
            attn_input = self.norm1(src)

        if self._add_reverse_edges:
            # Create reverse edges
            edge_sample_ids = edges[:, 0].repeat(2)
            edge_sources = torch.cat([edges[:, 1], edges[:, 2]])
            edge_targets = torch.cat([edges[:, 2], edges[:, 1]])
            edges = torch.stack((edge_sample_ids, edge_sources, edge_targets), dim=-1)
            edge_types = torch.cat([edge_types, edge_types + self._num_edge_types])

        queries, keys, values = self._compute_qkv(attn_input)
        src2 = self.self_attn(
            queries=queries,
            keys=keys,
            values=values,
            masked_elements=src_mask,
            edges=edges,
            edge_types=edge_types,
            dense_relations_kq=dense_relations_kq,
            dense_relations_kv=dense_relations_kv,
        )
        src2 = self._alpha1 * src2
        src = src + self.dropout1(src2)

        if post_self_att_hook is not None:
            src = post_self_att_hook(src, self._alpha1)

        if self._normalisation_mode == "postnorm":
            src = self.norm1(src)

        fc_input = src
        if self._normalisation_mode == "prenorm":
            fc_input = self.norm2(fc_input)
        src2 = self.linear2(self.dropout(self.activation(self.linear1(fc_input))))
        src2 = self._alpha2 * src2
        src = src + self.dropout2(src2)
        if self._normalisation_mode == "postnorm":
            src = self.norm1(src)
        return src


class RelationalTransformerDecoderLayer(nn.Module):
    def __init__(
        self,
        d_model: int,
        key_query_dimension: int,
        value_dimension: int,
        num_heads: int,
        num_self_edge_types: int,
        num_edge_types_to_encoder: int,
        add_reverse_edges: bool = True,
        dim_feedforward: int = 2048,
        dropout_rate: float = 0.1,
        activation: str = "relu",
        use_edge_value_biases: bool = False,
        edge_attention_bias_is_scalar: bool = False,
        rezero_mode: Literal["off", "scalar", "vector", "scalar-tied"] = "off",
        normalisation_mode: Literal["off", "prenorm", "postnorm"] = "postnorm",
    ):
        super().__init__()
        self.encoder = RelationalTransformerEncoderLayer(
            d_model=d_model,
            key_query_dimension=key_query_dimension,
            value_dimension=value_dimension,
            num_heads=num_heads,
            num_edge_types=2 * num_self_edge_types if add_reverse_edges else num_self_edge_types,
            add_reverse_edges=add_reverse_edges,
            dim_feedforward=dim_feedforward,
            dropout_rate=dropout_rate,
            activation=activation,
            use_edge_value_biases=use_edge_value_biases,
            edge_attention_bias_is_scalar=edge_attention_bias_is_scalar,
            rezero_mode=rezero_mode,
            normalisation_mode=normalisation_mode,
        )
        self.dropout = nn.Dropout(dropout_rate)
        self._key_query_dim = key_query_dimension
        self._value_dim = value_dimension
        self._multi_head_att_transforms = nn.Linear(
            in_features=d_model,
            out_features=num_heads * (key_query_dimension + value_dimension),
            bias=False,
        )
        self._query_transforms = nn.Linear(
            in_features=d_model, out_features=num_heads * key_query_dimension, bias=False
        )
        self.multihead_attn = RelationalMultiheadAttention(
            num_heads=num_heads,
            output_dimension=d_model,
            dropout_rate=dropout_rate,
            num_edge_types=2 * num_edge_types_to_encoder
            if add_reverse_edges
            else num_edge_types_to_encoder,
            key_query_dimension=key_query_dimension,
            value_dimension=value_dimension,
            use_edge_value_biases=use_edge_value_biases,
            edge_attention_bias_is_scalar=edge_attention_bias_is_scalar,
        )

    def forward(
        self,
        tgt: torch.Tensor,
        memory: torch.Tensor,
        tgt_mask: torch.Tensor,
        memory_mask: torch.Tensor,
        self_edges: torch.Tensor,
        self_edge_types: torch.Tensor,
        encoder_edges: torch.Tensor,
        encoder_edge_types: torch.Tensor,
        dense_self_relations_kq: Optional[torch.Tensor] = None,
        dense_self_relations_kv: Optional[torch.Tensor] = None,
        dense_encoder_relations_kq: Optional[torch.Tensor] = None,
        dense_encoder_relations_kv: Optional[torch.Tensor] = None,
    ):
        """
        :param tgt: A [batch_size, seq_len, D] tensor.
        :param memory: A [batch_size, mem_len, D] tensor.
        :param tgt_mask: A [batch_size, seq_len] bool tensor.  `True` values are those
            that should be masked (no attention paid).
        :param memory_mask: A [batch_size, mem_len] bool tensor.  `True` values are those
            that should be masked (no attention paid).
        :param self_edges: [num_self_edges, 3] each row has the form (batch_idx, source_idx, target_idx)
        :param self_edge_types: [num_self_edges] of integers from 0..num_self_edges
        :param encoder_edges: [num_enc_edges, 3] each row has the form (batch_idx, source_idx, target_idx)
            Note: target_idx refers to elements in the memory.
        :param encoder_edge_types: [num_enc_edges] of integers from 0..num_enc_edges
        :param dense_self_relations_kq: [batch_size, seq_len, seq_len, num_heads]
        :param dense_self_relations_kv: [batch_size, seq_len, seq_len, num_heads, value_dimension]
        :param dense_encoder_relations_kq: [batch_size, seq_len, mem_len, num_heads]
        :param dense_encoder_relations_kv: [batch_size, seq_len, mem_len, num_heads, value_dimension]

        :return:  [batch_size, seq_len, H]
        """

        def callback(src: torch.Tensor, rezero_alpha: Union[float, torch.Tensor]) -> torch.Tensor:
            kv = self._multi_head_att_transforms(memory).reshape(
                memory.shape[0], memory.shape[1], self.multihead_attn._num_heads, -1
            )
            keys, values = (
                kv[:, :, :, : self._key_query_dim],
                kv[:, :, :, self._key_query_dim :],
            )
            queries = self._query_transforms(src).reshape(
                src.shape[0], src.shape[1], self.multihead_attn.num_heads, -1
            )

            src2 = self.multihead_attn(
                queries=queries,
                keys=keys,
                values=values,
                masked_elements=memory_mask,
                edges=encoder_edges,
                edge_types=encoder_edge_types,
                dense_relations_kq=dense_encoder_relations_kq,
                dense_relations_kv=dense_encoder_relations_kv,
            )

            return src + self.dropout(rezero_alpha * src2)

        return self.encoder(
            tgt,
            tgt_mask,
            edges=self_edges,
            edge_types=self_edge_types,
            dense_relations_kq=dense_self_relations_kq,
            dense_relations_kv=dense_self_relations_kv,
            post_self_att_hook=callback,
        )
